{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "42a4c38d-110d-4c92-bb5d-968861b1bf32",
   "metadata": {},
   "source": [
    "# Makefiles\n",
    "\n",
    "\n",
    "\n",
    "## What is a Makefile\n",
    "\n",
    "A makefile allows you to atomatize the compilation process by following\n",
    "given rules, and for large projects this could actually speed up\n",
    "significantly the compilation process. You can also use a Makefile for\n",
    "latex projects and other related task where you want to reload and\n",
    "create automatically an updated version of a document when you have\n",
    "updated a small part. A rule in a makefile is written as\n",
    "\n",
    "```makefile\n",
    "target : dependencies separated by space\n",
    "    rule to build target\n",
    "#sumupdown.x : sumupdown.cpp\n",
    "#   g++ sumupdown.cpp -o sumupdown.x\n",
    "```\n",
    "\n",
    "It is important to note that the beginning of the rule is indented with\n",
    "**real tab**, not with spaces.\n",
    "\n",
    "\n",
    "Some modern tools/alternatives\n",
    "\n",
    "-   `remake` : <https://remake.readthedocs.io/en/latest/index.html>\n",
    "-   `just` : <https://just.systems/>,\n",
    "    <https://cheatography.com/linux-china/cheat-sheets/justfile/>\n",
    "-   `cmake`: <https://cmake.org/>,\n",
    "    <https://crascit.com/professional-cmake/>\n",
    "-   `xmake`: <https://xmake.io/#/>\n",
    "-   `taskfile`: <https://taskfile.dev/>\n",
    "-   `scons`: <https://scons.org/>\n",
    "-   `ninja, bazel, ...`\n",
    "\n",
    "\n",
    "**Note:** For more information, please see:\n",
    "\n",
    "-   [Gnu make\n",
    "    tutorial](https://www.gnu.org/software/make/manual/html_node/index.html)\n",
    "-   [Software Carpentry\n",
    "    Tutorial](https://swcarpentry.github.io/make-novice/) .\n",
    "-   <https://kbroman.org/minimal_make/>\n",
    "-   <https://swcarpentry.github.io/make-novice/>\n",
    "-   First linux version:\n",
    "    <https://elixir.bootlin.com/linux/0.01/source/kernel/Makefile> (more\n",
    "    info at <https://lwn.net/SubscriberLink/928581/841b747332791ac4/>)\n",
    "- <https://www.gnu.org/software/make/manual/html_node/Automatic-Prerequisites.html>\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2bd69145",
   "metadata": {},
   "source": [
    "\n",
    "## Application example: Final figure creation\n",
    "\n",
    "Imagine that you want to automatize the creation of the figure of the\n",
    "sumup and sumdown exercise from the numerical errors section. In this\n",
    "case you should think about the dependencies: the latest figure,\n",
    "`sumupdown.pdf` will depend on the a given gnuplot (or python) script\n",
    "(`plot.gp`) that read some data from a given file (`data.txt`), so any\n",
    "change on those files should trigger and update on the figure.\n",
    "Furhtermore, the data file will actually depend on the source\n",
    "(`sumupdown.cpp`) and its related executable (`sumupdown.x`), since any\n",
    "change there may imply new data. And actually the executable depends\n",
    "directly on the source code. We could express this dependencies as\n",
    "\n",
    "Therefore we have:\n",
    "- The executable depends on the source code, if it changes, then it must be recompiled\n",
    "  ```bash\n",
    "    sumupdown.x: sumupdown.cpp\n",
    "        g++ sumupdown.cpp -o sumupdown.x\n",
    "  ```\n",
    "  \n",
    "  This is the typical usage, to update executables. \n",
    "\n",
    "- The data depends on the executable: if it changes then the data must be generated again\n",
    "  ```bash\n",
    "    data.txt: sumupdown.x\n",
    "        ./sumupdown.x > data.txt\n",
    "  ```\n",
    "  \n",
    "  Changes in the executable (or changes in the source code) will trigger this rule.\n",
    "\n",
    "- The final figure depends on updates on the plotting script and the data\n",
    "  ```bash\n",
    "  fig.pdf: script.gp data.txt\n",
    "      gnuplot script.gp\n",
    "  ```\n",
    "  \n",
    "  By the dependency tree formed, changes in the source, executable, or data, will trigger this.\n",
    "\n",
    "\n",
    "Our final simple makefile will look like\n",
    "\n",
    "``` bash\n",
    "fig.pdf: script.gp data.txt\n",
    "    gnuplot script.gp\n",
    "\n",
    "data.txt: sumupdown.x\n",
    "    ./sumupdown.x > data.txt\n",
    "\n",
    "sumupdown.x: sumupdown.cpp\n",
    "    g++ sumupdown.cpp -o sumupdown.x\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e597ec0d",
   "metadata": {},
   "source": [
    "\n",
    "The actual order is not important. Every time you run make it will check\n",
    "if any target is outdated and run the command needed. The follwing code\n",
    "shows a simple solution for the sum up and down. You need to actually implement the functions, parametrize the data type, and also read the limit from the command line. "
   ]
  },
  {
   "cell_type": "markdown",
   "id": "19250f4e",
   "metadata": {},
   "source": [
    "\n",
    "```cpp\n",
    "#include <iostream>\n",
    "#include <cmath>\n",
    "\n",
    "float sumup(int nterms);\n",
    "float sumdown(int nterms);\n",
    "\n",
    "int main(int argc, char **argv)\n",
    "{\n",
    "  std::cout.precision(6); std::cout.setf(std::ios::scientific);\n",
    "  for (int ii = 1; ii <= 10000; ++ii) {\n",
    "    double sum1 = sumup(ii);\n",
    "    double sum2 = sumdown(ii);\n",
    "    std::cout << ii << \"\\t\" << sum1 << \"\\t\" << sum2\n",
    "              << \"\\t\" << std::fabs(sum1-sum2)/sum2 << \"\\n\";\n",
    "  }\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "\n",
    "\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f4c6d4e4",
   "metadata": {},
   "source": [
    "\n",
    "And this is a simple [gnuplot](http://www.gnuplot.info/) script example (you can also use [matplotlib](https://matplotlib.org/) or any tool that can be called from the command line)\n",
    "\n",
    "```bash\n",
    "set term pdf\n",
    "set out \"sumupdown.pdf\"\n",
    "# set xlabel \"Nterms\"\n",
    "# set ylabel \"Percentual difference\"\n",
    "plot 'data.txt' u 1:4 w lp pt 4\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2afdb8be",
   "metadata": {},
   "source": [
    "\n",
    "Now just run `make` and see what happens. Then uncomment, for example,\n",
    "the commented lines inside the gnuplot script and run make again. Check\n",
    "that make just reruns the needed rule, it does not recompile the whole\n",
    "program.\n",
    "\n",
    "Actually, the makefile can be simplified and generalized using variables\n",
    "\n",
    "``` bash\n",
    "fig.pdf: script.gp data.txt\n",
    "    gnuplot $^\n",
    "\n",
    "data.txt: sumupdown.x\n",
    "    ./$^ > $@\n",
    "\n",
    "# Generic rule to compile any cpp into its corresponding .x \n",
    "%.x: %.cpp\n",
    "    g++ $^ -o $@\n",
    "```\n",
    "\n",
    "Explanation:\n",
    "\n",
    "- `$^` : This is the first dependency\n",
    "- `$@` : Thisis the target\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2a41c6f8",
   "metadata": {},
   "source": [
    "\n",
    "## Application example: many files to compile\n",
    "\n",
    "Now, let's try to test our functions. To do so, we need another main function to test them, to separate our actual use from the tests. Therefore, it is better to put the functions declarations and implementations into their own files. You will now have three files:\n",
    "- `functions.h`: It will have the functions declarations. Add an inclusion guard with `pragma` or `define`.\n",
    "- `functions.cpp`: It will have the functions implementations (and will include `functions.h`). \n",
    "- `main.cpp`: This has the main function. Includes the headers `functions.h`. You can create as many main files as you want, one for using the function another for testing etc.\n",
    "\n",
    "Your first task is to move the declarations into `functions.h`, and the implementations into `functions.cpp`. Remember to include `functions.h` both into the implementations and then into the main file.\n",
    "\n",
    "The second task is to update the makefile to compile the whole thing. The manual command is"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "038da4ab",
   "metadata": {},
   "source": [
    "``` bash\n",
    "# this creates the object functions.o\n",
    "g++ -c functions.cpp\n",
    "# this creates the object main.o\n",
    "g++ -c main.cpp\n",
    "# this links with main and creates executable main.x\n",
    "g++ main.o functions.o -o main.x\n",
    "```\n",
    "What could be the corresponding make rule to build the objects and the executable? here is an example for the later\n",
    "\n",
    "```bash\n",
    "main.x : main.o functions.o\n",
    "    g++ -o main.x main.o functions.o\n",
    "```\n",
    "Test this with the old main function. **Where are the rules for creating the object files?** The objects are created automatically. These is done through something called automatic rules.\n",
    "\n",
    "To test the usefulness of this:\n",
    "- Change a bit the functions.cpp file and recompile with make. Are all objects updated?\n",
    "- Create another main file called test.cpp which test something about your functions (like calling them with negative limits). Add the rule to the makefile, so you can just write `make test` to test the functions.\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c40ff797",
   "metadata": {},
   "source": [
    "\n",
    "After seeing this advantages, let's try to generalize and use more\n",
    "useful syntax for make. Fir instance, you can specify variables, which\n",
    "can be overridden on the command line. Something like\n",
    "\n",
    "``` makefile\n",
    "CXX=g++\n",
    "CXXFLAGS=-I.\n",
    "\n",
    "main.x: foo.o bar.o\n",
    "  $(CXX) $(CXXFLAGS) -o main.x main.cpp foo.o bar.o\n",
    "```\n",
    "\n",
    "Compile again and test. Everything should work.\n",
    "\n",
    "Now, we will specify a generic rule to create the .o objects, and also\n",
    "the name of the objects in a variable. Furthermore, we use `$@` to\n",
    "specify the name of the target, and `$^` which symbolizes all the items\n",
    "in the dependency list\n",
    "\n",
    "``` makefile\n",
    "CXX = g++\n",
    "CXXFLAGS = -I.\n",
    "OBJ = foo.o bar.o\n",
    "\n",
    "main.x: $(OBJ)\n",
    "  $(CXX) $(CXXFLAGS) -o $@ $^ main.cpp\n",
    "```\n",
    "\n",
    "Save and run it, you will still get the same but the makefile is\n",
    "becoming more generic, and therefore more useful.\n",
    "\n",
    "Now let's specify the generic rule to create the `.o` files, and also\n",
    "specify that if we change the `.h` files then a recompilation is needed\n",
    "(for now our makefile only detect changes in `.cpp` files):\n",
    "\n",
    "``` makefile\n",
    "CXX = g++\n",
    "CXXFLAGS = -I.\n",
    "OBJ = foo.o bar.o\n",
    "DEPS = foo.h bar.h\n",
    "\n",
    "%.o: %.cpp $(DEPS)\n",
    "  $(CXX) -c -o $@ $< $(CXXFLAGS)\n",
    "\n",
    "main.x: $(OBJ)\n",
    "  $(CXX) $(CXXFLAGS) -o $@ $^ main.cpp\n",
    "```\n",
    "\n",
    "Where we have specify the dependency on `DEPS`, and also we are using a\n",
    "generic rule to tell how to create any `.o` file (exemplified with the\n",
    "generic `%.o` symbol), from any corresponding `%.c` file, and `$<` means\n",
    "the first item on the deps list. Again, save it and run it.\n",
    "\n",
    "You can rules to clean the directory (to erase object files, for\n",
    "example), to add libs, to put the headers on another directory, etc. For\n",
    "example, the following adds a phony rule (which does not create any\n",
    "target) to clean the directory\n",
    "\n",
    "``` makefile\n",
    "CXX = g++\n",
    "CXXFLAGS = -I.\n",
    "OBJ = foo.o bar.o\n",
    "DEPS = foo.h bar.h\n",
    "\n",
    "%.o: %.cpp $(DEPS)\n",
    "  $(CXX) -c -o $@ $< $(CXXFLAGS)\n",
    "\n",
    "main.x: $(OBJ)\n",
    "  $(CXX) $(CXXFLAGS) -o $@ $^ main.cpp\n",
    "\n",
    ".PHONY: clean\n",
    "clean:\n",
    "  rm -f *.o *~ *.x\n",
    "```\n",
    "\n",
    "In order to run this rule, you need to write\n",
    "\n",
    "``` shell\n",
    "make -f makefile-v5 clean\n",
    "```\n",
    "\n",
    "And, finally, a more complete example with comments and new syntax,\n",
    "which you can adapt for your own needs\n",
    "\n",
    "``` makefile\n",
    "CXX = g++\n",
    "CXXFLAGS = -I.\n",
    "LDFLAGS =\n",
    "SOURCES = main.cpp foo.cpp bar.cpp\n",
    "OBJ = $(SOURCES:.cpp=.o) # extracts automatically the objects names\n",
    "DEPS = foo.h bar.h\n",
    "\n",
    "all : main.x $(SOURCES) $(DEPS) # this is the default target\n",
    "\n",
    "main.x: $(OBJ)\n",
    "  @echo \"Creating main executable ...\"\n",
    "  $(CXX) $(CXXFLAGS) -o $@ $^ $(LDFLAGS)\n",
    "\n",
    ".cpp.o:\n",
    "  $(CXX) -c -o $@ $< $(CXXFLAGS)\n",
    "\n",
    ".PHONY: clean\n",
    "clean:\n",
    "  rm -f *.o *~ *.x\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "d7912ce0",
   "metadata": {},
   "source": [
    "\n",
    "## Exercises\n",
    "\n",
    "1.  Take the codes for overflow and underflow and automatize the\n",
    "    detection of overflow with the standard function `isinf`.\n",
    "2.  When they are detecting overflow automatically, stopping and\n",
    "    printing where the overflow occurs, put them in separated files\n",
    "    (`.h` and `.cpp`).\n",
    "3.  Then create a main file which calls those functions.\n",
    "4.  Now write a makefile to automatize the compilation. Use the latest\n",
    "    version of the example Makefile.\n",
    "5.  When everything is working send it to assigned repository.\n",
    "6.  Do the same now including the code which computes the machine eps,\n",
    "    using its own `.h` and `.cpp` .\n",
    "7.  If you use latex, how will you write a Makefile for that? Well, at\n",
    "    the end you can use `latexmk`, of course. Try to think on other uses\n",
    "    for make (like automatic zipping and mailing of some important file,\n",
    "    )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2524a62d",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
